﻿using System.Collections.Generic;
using System.Collections;
using System;
using System.Windows.Forms;

namespace TSPForm
{
    public class CACO
    {
        //const int N = 31;   //城市数目
        static int N = 29;
        static int M = 25;   //蚂蚁数目
        double inittao = 1; //初始时刻，信息素值
        int Nc = 0;        //当前迭代次数 

        PictureBox pictureBoxTSP; //show the city and road of TSP 
        Label labelCurrentLength; //show the current length of TSP
        FrmTSPSolver frmTSPSolver;
        /// <summary>
        /// N个城市的横坐标
        /// </summary>
        private double[] x;

        /// <summary>
        /// N个城市的纵坐标
        /// </summary>
        private double[] y;

        /// <summary>
        /// 城市距离矩阵(d_ij)
        /// </summary>
        // private double[,] distance = new double[N, N];
        private double[,] distance;
        /// <summary>
        /// 信息素矩阵(tau_ij)
        /// </summary>
        //  private double[,] tao = new double[N, N];
        private double[,] tao;
        /// <summary>
        /// 启发函数矩阵(eta_ij)
        /// </summary>
        private double[,] eta;

        /// <summary>
        /// 信息素增量矩阵(detatao_ij)
        /// </summary>
        private double[,] detatao;

        /// <summary>
        /// 信息启发式因子，表示轨迹的相对重要性
        /// </summary>
        private double alpha;

        /// <summary>
        /// 期望式启发因子，表示能见度的相对重要性
        /// </summary>
        private double beta;

        /// <summary>
        /// 表示信息挥发系数，（1-rho）表示信息残留因子，通常[0 <= rho <= 1]
        /// </summary>
        private double rho;

        /// <summary>
        /// 信息素强度
        /// </summary>
        private double Q;

        /// <summary>
        /// 最大迭代次数
        /// </summary>
        private int Nc_max;

        /// <summary>
        /// 第k只蚂蚁的禁忌表，记录蚂蚁k走过的城市
        /// </summary>
        private double[,] tabu;

        /// <summary>
        /// 蚂蚁k的下一步可选城市表
        /// </summary>
        private int[,] no_tabu;

        /// <summary>
        /// 记录蚂蚁k走过的城市
        /// </summary>
        private int[,] route;

        /// <summary>
        /// 存放最佳路径上所经过的各个城市
        /// </summary>
        private int[] bestRoute;

        /// <summary>
        /// 存放M只蚂蚁的最短路径解（即最短路径长度）
        /// </summary>
        private double[] solution;

        /// <summary>
        /// 初始化最优解路径为无穷大
        /// </summary>
        private double bestSolution = double.PositiveInfinity;

        /// <summary>
        /// 蚂蚁从城市i到城市j选择的概率阈值
        /// </summary>
        private double dPijThreshold;

        /// <summary>
        /// Class construct funtion  of CACO
        /// </summary>
        /// <param name="m">ant Count;</param>
        /// <param name="n">City count</param>
        /// <param name="alpha">Heuristic Coefficient</param>
        /// <param name="beta">Expectation Coefficient</param>
        /// <param name="rho">Information Volatilization:0<=ρ<1</param>
        /// <param name="Q">Pheromone Intensity</param>
        /// <param name="Nc_max">Iteration Count</param>
        /// <param name="x">City coordinate Array of X axis</param>
        /// <param name="y">City coordinate Array of Y axis</param>
        /// <param name="C">Initial Pheromone Const for every Edge</param>
        /// <param name="bestRoute">Record the best root path </param>
        /// <param name="frmTSPSolver">parent form hand </param>
        //  public CACO(int m, int n, double alpha, double beta, double rho, double Q, int Nc_max, double[] x, double[] y, double C, ref PictureBox pictureBoxTSP, ref Label labelCurrentLength, ref int[] bestRoute, FrmTSPSolver frmTSPSolver)
        public CACO(int m, int n, double alpha, double beta, double rho, double Q, int Nc_max, double[] x, double[] y, double C, ref int[] bestRoute, FrmTSPSolver frmTSPSolver)
        {
            M = m;
            N = n;
            this.alpha = alpha;
            this.beta = beta;
            this.rho = rho;
            this.Q = Q;
            this.Nc_max = Nc_max;
            this.x = x;
            this.y = y;
            inittao = C; // 2014.11.19-16:20
            // this.pictureBoxTSP = pictureBoxTSP;
            //  this.labelCurrentLength = labelCurrentLength;
            this.frmTSPSolver = frmTSPSolver;
            distance = new double[N, N];
            tao = new double[N, N];
            eta = new double[N, N];
            detatao = new double[N, N];
            tabu = new double[M, N];
            no_tabu = new int[M, N];
            route = new int[M, N];
            bestRoute = new int[N];
            solution = new double[M];
            this.bestRoute = bestRoute;
        }

        /// <summary>
        /// Initial information  of the ACO algorithm 
        /// </summary>
        public void Init()
        {
            #region  Algorithm Initiation

            #region Initial the distance matrix of every two cities

            for (int i = 0; i < N; i++)
            {
                frmTSPSolver.setProgressBarMaxMin(0, N);
                for (int j = 0; j < N; j++)
                {
                    distance[i, j] = Math.Sqrt((x[i] - x[j]) * (x[i] - x[j]) + (y[i] - y[j]) * (y[i] - y[j]));
                    frmTSPSolver.setProgressBar(j + 1);
                }
            }

            #endregion

            #region Initial the pheromone  matrix and eta(heuristic function )
            frmTSPSolver.setProgressBarMaxMin(0, N);
            for (int i = 0; i < N; i++)
            {
                for (int j = 0; j < N; j++)
                {
                    tao[i, j] = inittao;  //Initial the pheromone  matrix 初始化信息素矩阵 
                    if (i != j)
                    {
                        eta[i, j] = 1 / distance[i, j];  //Initail the heuristic gene  计算启发因子
                    }
                }
                frmTSPSolver.setProgressBar(i + 1);
            }
            #endregion

            #region Initail the pheromone increment matrix  初始化信息素增量矩阵
            frmTSPSolver.setProgressBarMaxMin(0, N);
            for (int i = 0; i < N; i++)
            {
                for (int j = 0; j < N; j++)
                {
                    detatao[i, j] = 0.0;
                }
                frmTSPSolver.setProgressBar(i + 1);
            }
            #endregion


            #region Initial the flag information
            frmTSPSolver.setProgressBarMaxMin(0, M);
            for (int k = 0; k < M; k++)
            {
                for (int j = 0; j < N; j++)
                {
                    route[k, j] = -1;    //将蚂蚁k已走过的城市置-1,表示还都没走过
                    tabu[k, j] = 0;      //将蚂蚁k的禁忌表置0，表示当前所有城市都可以访问
                }
                frmTSPSolver.setProgressBar(k + 1);
            }
            #endregion

            #region Initial ant start  location
            for (int k = 0; k < M; k++)
            {
                if (CAppParamter.iStartCityNo != -1) //Using the index set by user
                {
                    route[k, 0] = CAppParamter.iStartCityNo - 1;
                    tabu[k, route[k, 0]] = 1; 
                }
                else
                {
                    for (int j = 0; j < N; j++)
                    {
                        route[k, 0] = k % N;  //设置蚂蚁k的初始城市
                        tabu[k, route[k, 0]] = 1;  //将蚂蚁k的出生城市放入禁忌表中，值访问位为1
                    }
                }

            }
            #endregion

            #endregion
        }

        /// <summary>
        /// Find the Best path of TSP using ACO algorithm  
        /// </summary>
        public void FindOptimalPath()
        {
            do
            {
                int s = 1;  //遍历的城市个数，等于N时表示完成一次遍历
                double p_ij = 0;   //蚂蚁k从城市i转移到城市j的概率
                double partsum = 0;  //p_ij计算式中的分母
                double random_num = 0;  //[0,1]之间的随机数
                Random rand = new Random();
                frmTSPSolver.setProgressBarMaxMin(0, N);
                while (s < N)     //开始下一次转移
                {
                    #region 开始蚂蚁的运动
                    for (int k = 0; k < M; k++)    //每只蚂蚁都做出动作，本应是同时进行，这里采用每只蚂蚁依次动作，效果一样
                    {
                        p_ij = 0;  //转移概率
                        partsum = 0;

                        random_num = rand.Next(3000) / 3000.0;  //除以3000.0是为了使生成的随机数在[0,1]内，同时使其变为double类型
                        dPijThreshold = random_num; // 2014.11.19-17:17 目前使用随机概率
                        for (int j = 0; j < N; j++)
                        {
                            if (tabu[k, j] == 0)    //选出未访问的城市
                            {
                                partsum += Math.Pow(tao[route[k, s - 1], j], alpha) * Math.Pow(eta[route[k, s - 1], j], beta);  //转移概率p的分母
                                //route[k,s-1] 表示蚂蚁k在s-1步走到的城市，s==1表示其出发的城市
                                //tao[route[k,s-1],j] 表示蚂蚁k从刚访问过的城市i到下一个未访问的城市j的信息量
                            }
                        }
                        for (int j = 0; j < N; j++)
                        {
                            if (tabu[k, j] == 0)
                            {
                                //当只剩下一个城市未被访问时 p_ij == 1
                                p_ij += Math.Pow(tao[route[k, s - 1], j], alpha) * Math.Pow(eta[route[k, s - 1], j], beta) / partsum;  //计算转移概率
                                if (p_ij > dPijThreshold)   //此时的 j 即为蚂蚁k下一步转移的城市
                                {
                                    tabu[k, j] = 1;   //修改禁忌表，表示城市j已经走过
                                    route[k, s] = j;  //修改已访问的路径表
                                    break;
                                }
                            }
                        }
                    }
                    #endregion
                    frmTSPSolver.setProgressBar(s);
                    s++;
                }

                //选出最佳路径并保存
                #region N次循环后，所有蚂蚁的禁忌表都填满了，都找到了一个解,接下来要做的就是，挑选出最优解并保存最佳路径，更新各路径上的信息量
                int iMinLengAnt = 0;
                frmTSPSolver.setProgressBarMaxMin(0, M);
                for (int k = 0; k < M; k++)
                {
                    solution[k] = EvaluateSolution(k);  //计算蚂蚁k所经过路径的长度，并保存在数组solution[]中
                    if (solution[k] < bestSolution)
                    {
                        bestSolution = solution[k];     //更新最短路径长度
                        iMinLengAnt = k;
                        //for (int i = 0; i < N; i++)
                        //{
                        //    bestRoute[i] = route[k, i]; //更新最短路径
                        //}
                    }
                    frmTSPSolver.setProgressBar(k + 1);
                }
                frmTSPSolver.setProgressBarMaxMin(0, N);
                for (int i = 0; i < N; i++)
                {
                    bestRoute[i] = route[iMinLengAnt, i]; //更新最短路径
                    frmTSPSolver.setProgressBar(i + 1);
                }
                #endregion

                #region 计算各路径上的信息素增量
                for (int k = 0; k < M; k++)
                {
                    frmTSPSolver.setProgressBarMaxMin(0, N - 1);
                    for (int j = 0; j < N - 1; j++)
                    {
                        detatao[route[k, j], route[k, j + 1]] += Q / solution[k];    //信息素增量可以累积，由经过该路段的蚂蚁数目决定
                        frmTSPSolver.setProgressBar(j + 1);
                    }
                    detatao[route[k, N - 1], route[k, 0]] += Q / solution[k];
                }
                #endregion

                #region 更新各路径上的信息量
                for (int i = 0; i < N; i++)
                {
                    frmTSPSolver.setProgressBarMaxMin(0, N);
                    for (int j = 0; j < N; j++)
                    {
                        tao[i, j] = (1 - rho) * tao[i, j] + detatao[i, j];   //由基本蚁群算法的信息素公式计算
                        if (tao[i, j] < 0.00001)  //信息素下界，信息素过少不能较好的加深最佳解
                            tao[i, j] = 0.00001;
                        if (tao[i, j] > 20)       //信息素上界，信息素过多会淹没启发信息
                            tao[i, j] = 20;
                        frmTSPSolver.setProgressBar(j + 1);
                    }
                }
                #endregion

                #region 重置禁忌表和路径表，为下一次迭代做好准备
                frmTSPSolver.setProgressBarMaxMin(0, M);
                for (int k = 0; k < M; k++)
                {
                    for (int j = 1; j < N; j++)  //j从1开始，起始点城市没有清空
                    {
                        route[k, j] = -1;
                    }
                    for (int i = 0; i < N; i++)
                    {
                        tabu[k, i] = 0;
                    }
                    tabu[k, route[k, 0]] = 1;
                    frmTSPSolver.setProgressBar(k + 1);
                }
                #endregion

                frmTSPSolver.OutPutLength(Nc + 1, bestSolution);
                frmTSPSolver.DrawMinRoadLine(bestRoute);
            } while (++Nc < Nc_max);
        }

        /// <summary>
        /// 计算蚂蚁k所经过路径的总长度
        /// </summary>
        /// <param name="k">蚂蚁k</param>
        /// <returns>蚂蚁k经过路径总长</returns>
        private double EvaluateSolution(int k)
        {
            double dis = 0;
            for (int i = 0; i < N - 1; i++)
            {
                dis += distance[route[k, i], route[k, i + 1]];
            }
            dis += distance[route[k, N - 1], route[k, 0]];  //走到最后一个城市后又返回起始点
            return dis;
        }
    }
}
