﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace TSPForm
{
    public class CGreedyTSP
    {

        static int N = 100; //City Count 
        FrmTSPSolver frmTSPSolver;
        /// <summary>
        /// N个城市的横坐标
        /// </summary>
        private double[] x;

        /// <summary>
        /// N个城市的纵坐标
        /// </summary>
        private double[] y;

        /// <summary>
        /// 城市距离矩阵(d_ij)
        /// </summary>
        private double[,] distance;

        /// <summary>
        ///记录已走过的城市-禁忌表
        /// </summary>
        private double[] tabu;
        
        /// <summary>
        /// 记录蚂蚁k走过的城市
        /// </summary>
        private int[] route;
        /// <summary>
        /// 存放最佳路径上所经过的各个城市
        /// </summary>
        private int[] bestRoute;
        /// <summary>
        /// 初始化最优解路径为无穷大
        /// </summary>
        private double bestSolution = double.PositiveInfinity;

        /// <summary>
        ///  creterion of selection
        /// </summary>
        private double dNextThreshold;

        private int iStartCityNo; //Start city No of TSP 

        /// <summary>
        /// Construct function of class CGreedyTSP
        /// </summary>
        /// <param name="iCityCountN">Count of city </param>
        /// <param name="x">City coordinate Array of X axis</param>
        /// <param name="y">City coordinate Array of Y axis</param>
        /// <param name="iStartCityNo"></param>
        /// <param name="bestRoute">Record the best root path</param>
        /// <param name="frmTSPSolver">parent form hand</param>
        public CGreedyTSP(int iCityCountN, double[] x, double[] y, int iStartCityNo, ref int[] bestRoute, FrmTSPSolver frmTSPSolver)
        {
            N = iCityCountN;
            this.x = x;
            this.y = y;
            this.iStartCityNo = iStartCityNo;
            this.frmTSPSolver = frmTSPSolver;
            distance = new double[N, N];
            tabu = new double[N];
            route = new int[N];
            bestRoute = new int[N];
            this.bestRoute = bestRoute;
        }

        /// <summary>
        /// Initial information  of the Greedty algorithm 
        /// </summary>
        public void Init()
        {
            #region Initial the distance matrix of every two cities
            frmTSPSolver.setProgressBarMaxMin(0, N);
            for (int i = 0; i < N; i++)
            {
                for (int j = 0; j < N; j++)
                {
                    distance[i, j] = Math.Sqrt((x[i] - x[j]) * (x[i] - x[j]) + (y[i] - y[j]) * (y[i] - y[j]));
                }
                frmTSPSolver.setProgressBar(i + 1);
            }
            #endregion

            #region Initial the flag information
            frmTSPSolver.setProgressBarMaxMin(0, N);
            for (int j = 0; j < N; j++)
            {
                route[j] = -1;    //set the array value of all cities  to one which  indicate the relative city has not been trived
                tabu[j] = 0;      // set the relative array value of tabu city zero which indicate the city can be trived
                bestRoute[j] = 0;
                frmTSPSolver.setProgressBar(j + 1);
            }
            #endregion

            #region Initial start  location of TSP 
            route[0] = iStartCityNo - 1;//Record the Start city of TSP
            tabu[route[0]] = 1;   //将已经走过的城市放入禁忌表中，值访问位为1
            #endregion
        }


        /// <summary>
        /// Find the Best path of TSP using greedy algorithm 
        /// </summary>
        public void FindOptimalPath()
        {
            int s = 1; //the count of cities which have ben traved
            double dijMin = double.MaxValue;//initial the minimum value beteew two cities 
            int iCureentMinIndex = route[0];
            frmTSPSolver.setProgressBarMaxMin(0, N-1);
            while (s < N)
            {
                dijMin = double.MaxValue;
                for (int i = 0; i < N; i++)
                {
                    if (tabu[i] == 0 && distance[route[s - 1], i] < dijMin)
                    {
                        dijMin = distance[route[s - 1], i];
                        iCureentMinIndex = i;
                    }
                }
                frmTSPSolver.setProgressBar(s);
                tabu[iCureentMinIndex] = 1; //Change the tabu array in which the value  of city trived was set to 1
                route[s] = iCureentMinIndex; //Record current city index which was seleclted using the greedy algorithm 
                //bestRoute[s] = iCureentMinIndex;
                frmTSPSolver.OutPutLength(s, 0); 
                s++;       
            }
            bestSolution = EvaluateSolution();
            frmTSPSolver.OutPutLength(s, bestSolution); 
            frmTSPSolver.DrawMinRoadLine(route);
        }

        /// <summary>
        /// Calculate the minimum distance of the TSP road which is selected using the greedy algorithm 
        /// </summary>
        /// <returns>the minimum distance</returns>
        private double EvaluateSolution()
        {
            double dis = 0;
            for (int i = 0; i < N - 1; i++)
            {
                dis += distance[route[i], route[i + 1]];
            }
            dis += distance[route[N - 1], route[0]];  //add the distance beetwenn the last city and the start city 
            return dis;
        }
    }


}
