﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace TSPForm
{
    public partial class FrmTSPSolver : Form
    {
        string strTestDataPath;//File Path of Test Data
        FrmACO frmACO;//  form of Parameter set of ACO Algorithm
        FrmGreedy frmGreedy;// form of parameter set 
        FrmAbout frmAbout;
        double[] xCoordinateArray; //store X coordinate of city
        double[] yCoordinateArray;//store Ycoordinate of city
        Bitmap imageTemp;
        Graphics formGraphics;
        int[] bestRoute; //Record the best hints of city point 
        public FrmTSPSolver()
        {
            InitializeComponent();
        }

        private void oPenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            openFileDialogTestData.InitialDirectory = Application.StartupPath + @"\TSP_TestData";
            //Reset the Parameters
            CAppParamter.dXMax = 0;
            CAppParamter.dXMin = float.MaxValue;
            CAppParamter.dYMax = 0;
            CAppParamter.dYMin = float.MaxValue;
            CAppParamter.CurrentNc = 0;
            xCoordinateArray = null;
            yCoordinateArray = null;
            labelCurrentLength.Text = "";
            labelIterativeCount.Text = "";
            LoadTestData();
        }

        /// <summary>
        /// Load Test Data of TSP
        /// </summary>
        private void LoadTestData()
        {

            if (openFileDialogTestData.ShowDialog() == DialogResult.OK) //Read TSP File Successfully 
            {
                strTestDataPath = openFileDialogTestData.FileName;
                ReadTSPData(strTestDataPath);
                Thread thdDraw = new Thread(DrawCityVertex);
                thdDraw.IsBackground = true;
                thdDraw.Start();  //Draw Vertex of all cities ;
                // DrawCityVertex();

                // Enable the munu of TSP Solver
                greedyAlgorithmToolStripMenuItem.Enabled = true;
                ACOlgorithmToolStripMenuItem.Enabled = true;
            }
            else
            {
                // disable the munu of TSP Solver
                greedyAlgorithmToolStripMenuItem.Enabled = false;
                ACOlgorithmToolStripMenuItem.Enabled = false;
            }
        }


        /// <summary>
        /// Read The TSP File
        /// </summary>
        /// <param name="strFilePath">File Name of TSP Test Data</param>
        private bool ReadTSPData(string strFilePath)
        {
            try
            {
                StreamReader streamReaderTSP = new StreamReader(strFilePath);
                string strLineData;
                string[] strDataArray;
                int iFlag = 0;
                int iCount = 0;
                writeMsgToList("Begin reading  Test Data...", 0);
                while ((strLineData = streamReaderTSP.ReadLine()) != "EOF")
                {
                    if (iFlag == 1)//Read the Data and Store it
                    {
                        //store data
                        strDataArray = strLineData.Split(' ');
                        xCoordinateArray[iCount] = float.Parse(strDataArray[1]);
                        yCoordinateArray[iCount] = float.Parse(strDataArray[2]);
                        CAppParamter.dXMax = xCoordinateArray[iCount] > CAppParamter.dXMax ? xCoordinateArray[iCount] : CAppParamter.dXMax;
                        CAppParamter.dXMin = xCoordinateArray[iCount] < CAppParamter.dXMin ? xCoordinateArray[iCount] : CAppParamter.dXMin;

                        CAppParamter.dYMax = yCoordinateArray[iCount] > CAppParamter.dYMax ? yCoordinateArray[iCount] : CAppParamter.dYMax;
                        CAppParamter.dYMin = yCoordinateArray[iCount] < CAppParamter.dYMin ? yCoordinateArray[iCount] : CAppParamter.dYMin;
                        iCount++;
                        writeMsgToList("Read city " + iCount.ToString(), 0);
                    }
                    if (strLineData.Contains("NODE_COORD_SECTION")) //Data in Next Line 
                    {
                        iFlag = 1;
                    }
                    if (strLineData.Contains("NAME")) //Read File Name
                    {
                        strDataArray = strLineData.Split(':');
                        labelTestDataName.Text = strDataArray[1].Trim();
                        writeMsgToList("Read the Name of test data...", 0);
                    }
                    if (strLineData.Contains("Optimum")) //Read Optimum tour length 
                    {
                        int iLenthStartIndex;
                        iLenthStartIndex = strLineData.IndexOf("is") + 2;
                        labelOptimunLength.Text = strLineData.Substring(iLenthStartIndex).Trim();
                        writeMsgToList("Read the Optimum tour length...", 0);
                    }
                    if (strLineData.Contains("DIMENSION")) //Read Dimension
                    {
                        strDataArray = strLineData.Split(':');
                        labelDimension.Text = strDataArray[1].Trim();
                        CAppParamter.iCityCount = int.Parse(strDataArray[1].Trim());
                        progressBarSolver.Maximum = CAppParamter.iCityCount;
                        xCoordinateArray = new double[CAppParamter.iCityCount];
                        yCoordinateArray = new double[CAppParamter.iCityCount];
                        bestRoute = new int[CAppParamter.iCityCount];
                        //cityDistanceArray = new double[CAppParamter.iCityCount, CAppParamter.iCityCount];
                        writeMsgToList("Read the DIMENSION  of test data...", 0);
                    }
                }
                writeMsgToList("Data have read successfully! ", 0);
                streamReaderTSP.Close();
                streamReaderTSP.Dispose();
            }
            catch (System.Exception ex)
            {
                MessageBox.Show("Can't Open the TSP data,please check the File or try anain !/n" + ex.ToString());
                return false;
            }

            return true;
        }

        /// <summary>
        /// Draw the city Vertex
        /// </summary>
        private void DrawCityVertex()
        {
            writeMsgToList("Now ,Begin to Draw the City Vertex ... ", 0);
            RestPicBox();
            paintPoint();
        }


        /// <summary>
        /// Reset the pictrue of formGraphics
        /// </summary>
        private void RestPicBox()
        {
            pictureBoxTSP.Image = null;
            imageTemp = new Bitmap(pictureBoxTSP.Width, pictureBoxTSP.Height);
            formGraphics = Graphics.FromImage(imageTemp);
            formGraphics.Clear(Color.AliceBlue);
        }

        public delegate void drawLineInvoke(int[] CurrentBest);
        /// <summary>
        /// Draw Current minimum road
        /// </summary>
        /// <param name="line"></param>
        /// <param name="flag"></param>
        public void DrawMinRoadLine(int[] CurrentBest)
        {
            RestPicBox();
            paintPoint();//Repaint the City Point ;
            if (pictureBoxTSP.InvokeRequired)
            {
                drawLineInvoke _setInvoke = new drawLineInvoke(DrawMinRoadLine);
                pictureBoxTSP.BeginInvoke(_setInvoke, new object[] { CurrentBest });
            }
            else
            {
                int i;
                float xa, ya, xb, yb;
                Pen p = new Pen(Color.Orange, 1);//brush pen

                float fXY = (float)CAppParamter.fXY;
                float fPicboxWidth = pictureBoxTSP.Width;
                float fPicboxheight = pictureBoxTSP.Height;
                double dXscale, dYscale;
                dXscale = (fPicboxWidth - fXY) / (CAppParamter.dXMax - CAppParamter.dXMin);
                dYscale = (fPicboxheight - fXY) / (CAppParamter.dYMax - CAppParamter.dYMin);
                float fXscale = (float)dXscale;
                float fYscale = (float)dYscale;
                for (i = 1; i < CAppParamter.iCityCount; i++)
                {
                    xa = (float)(xCoordinateArray[CurrentBest[i - 1]] - CAppParamter.dXMin) * fXscale;
                    ya = (float)(yCoordinateArray[CurrentBest[i - 1]] - CAppParamter.dYMin) * fYscale;

                    xb = (float)(xCoordinateArray[CurrentBest[i]] - CAppParamter.dXMin) * fXscale;
                    yb = (float)(yCoordinateArray[CurrentBest[i]] - CAppParamter.dYMin) * fYscale;
                    formGraphics.DrawLine(p, new PointF(xa, ya), new PointF(xb, yb));//DrawLine
                }

                xa = (float)(xCoordinateArray[CurrentBest[0]] - CAppParamter.dXMin) * fXscale;
                ya = (float)(yCoordinateArray[CurrentBest[0]] - CAppParamter.dYMin) * fYscale;

                xb = (float)(xCoordinateArray[CurrentBest[CAppParamter.iCityCount - 1]] - CAppParamter.dXMin) * fXscale;
                yb = (float)(yCoordinateArray[CurrentBest[CAppParamter.iCityCount - 1]] - CAppParamter.dYMin) * fYscale;
                formGraphics.DrawLine(p, new PointF(xa, ya), new PointF(xb, yb));//Draw the line which 

                //Draw the start City and end City of the current minimum road
                Color startColor = Color.Green;
                Color endColor = Color.Yellow;
                Font myFont = new Font("Hacttenschweiler", 7);
                Brush greenBrush = Brushes.Green;

                Pen ph = new Pen(startColor, 2);
                Pen pf = new Pen(endColor, 2);
                formGraphics.FillEllipse(Brushes.Red, new RectangleF(xa, ya, 9, 9));
                formGraphics.DrawString("Start Point", myFont, greenBrush, new PointF(xa, ya + 5));

                formGraphics.FillEllipse(Brushes.Blue, new RectangleF(xb, yb, 9, 9));
                formGraphics.DrawString("End Point", myFont, Brushes.Green, new PointF(xb, yb + 5));
                pictureBoxTSP.Image = (Image)imageTemp;

            }
        }


        public delegate void paintPointInvoke();
        /// <summary>
        /// Draw the City Point
        /// </summary>
        public void paintPoint()
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new paintPointInvoke(paintPoint));
            }
            else
            {
                int i;
                float x1, y1;
                float fScaleRatio = 1F;
                System.Drawing.SolidBrush myBrush = new System.Drawing.SolidBrush(System.Drawing.Color.Red);//画刷
                Brush blackBrush = Brushes.Black;
                Font myFont1 = new Font("Hacttenschweiler", 7);

                float fXY = (float)CAppParamter.fXY;
                float fPicboxWidth = pictureBoxTSP.Width;
                float fPicboxheight = pictureBoxTSP.Height;

                double dXscale, dYscale;
                dXscale = (fPicboxWidth - fXY) / (CAppParamter.dXMax - CAppParamter.dXMin);
                dYscale = (fPicboxheight - fXY) / (CAppParamter.dYMax - CAppParamter.dYMin);
                float fXscale = (float)dXscale;
                float fYscale = (float)dYscale;
                //float fXscale = fPicboxWidth  / (CAppParamter.dXMax - CAppParamter.dXMin);
                //float fYscale = fPicboxheight  / (CAppParamter.dYMax - CAppParamter.dYMin);
                for (i = 0; i < CAppParamter.iCityCount; i++)
                {
                    //x1 = fXY + (xCoordinateArray[i] - CAppParamter.dXMin) * fXscale;
                    //y1 = (yCoordinateArray[i] - CAppParamter.dYMin) * fYscale + fXY;

                    x1 = (float)(xCoordinateArray[i] - CAppParamter.dXMin) * fXscale;
                    y1 = (float)(yCoordinateArray[i] - CAppParamter.dYMin) * fYscale;
                    formGraphics.FillEllipse(myBrush, new RectangleF(x1, y1, 5, 5));//画实心椭圆
                    formGraphics.DrawString((i + 1).ToString(), myFont1, blackBrush, new PointF(x1, y1));//写字
                    setProgressBar(i + 1);

                }
                pictureBoxTSP.Image = (Image)imageTemp;
            }
        }

        /// <summary>
        /// 写入操作记录提示
        /// </summary>
        /// <param name="strMsg">要写的消息内容</param>
        /// <param name="iInserIndex">插入到listview中位置索引号</param>
        public void writeMsgToList(string strMsg, int iInserIndex)
        {
            writeMsgToList(listViewTipInfo, strMsg, iInserIndex);
        }

        /// <summary>
        ///将具体的消息显示在相应的listView中
        /// </summary>
        /// <param name="listViewTemp"> 显示的listview</param>
        /// <param name="strMsg">显示的消息实体</param>
        /// <param name="iInserIndex">0:表示插入到第一行，1表示插入到第二行</param>
        public void writeMsgToList(ListView listViewTemp1, string strMsg1, int iInserIndex1)
        {
            if (listViewTemp1.InvokeRequired)
            {
                Action<ListView, string, int> writeLstInto = delegate(ListView listViewTemp, string strMsg, int iInserIndex)
                {
                    //如果当前listView中items 多余200条则清除全部
                    if (listViewTemp.Items.Count >= 2000)
                    {
                        listViewTemp.Items.Clear();
                    }

                    //消息时间
                    string strMsgTime;
                    strMsgTime = DateTime.Now.ToString("HH:mm:ss");

                    listViewTemp.BeginUpdate();//避免屏幕闪动
                    ListViewItem itemTemp = new ListViewItem(strMsgTime + ":" + strMsg);
                    listViewTemp.Items.Insert(iInserIndex, itemTemp);
                    listViewTemp.EndUpdate();
                };
                listViewTemp1.BeginInvoke(writeLstInto, listViewTemp1, strMsg1, iInserIndex1);
            }
            else
            {
                //如果当前listView中items 多余200条则清除全部
                if (listViewTemp1.Items.Count >= 2000)
                {
                    listViewTemp1.Items.Clear();
                }
                //消息时间
                string strMsgTime;
                strMsgTime = DateTime.Now.ToString("HH:mm:ss");

                listViewTemp1.BeginUpdate();//避免屏幕闪动
                ListViewItem itemTemp = new ListViewItem(strMsgTime + ":" + strMsg1);
                listViewTemp1.Items.Insert(iInserIndex1, itemTemp);
                listViewTemp1.EndUpdate();
            }
        }



        /// <summary>
        /// output current minumun lenth of tsp
        /// </summary>
        /// <param name="iRound">current iteration  of TSP</param>
        /// <param name="dCurrentMinLength"> current minumun lenth</param>
        public void OutPutLength(int iRound, double dCurrentMinLength)
        {
            if (labelCurrentLength.InvokeRequired)
            {
                Action<double> actionWriteLenth = delegate(double dLength)
                {
                    labelCurrentLength.Text = dLength == 0 ? "" : dLength.ToString("F0");
                };
                labelCurrentLength.BeginInvoke(actionWriteLenth, dCurrentMinLength);
            }
            else
            {
                labelCurrentLength.Text = dCurrentMinLength == 0 ? "" : dCurrentMinLength.ToString("F0");
            }

            if (labelIterativeCount.InvokeRequired)
            {
                Action<int> actionWriteCount = delegate(int iCount)
                {
                    labelIterativeCount.Text = iCount.ToString();
                };
                labelIterativeCount.BeginInvoke(actionWriteCount, iRound);
            }
            else
            {
                labelIterativeCount.Text = iRound.ToString();
            }
        }


        /// <summary>
        /// Open the Parameters set form of ACO
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ACOlgorithmToolStripMenuItem_Click(object sender, EventArgs e)
        {
            labelCurrentLength.Text = "";
            labelIterativeCount.Text = "";
            labelAlgorithmType.Text = "ACO Algorithm progress...";
            if (frmACO == null || frmACO.IsDisposed)
            {
                frmACO = new FrmACO(this);
                frmACO.ShowDialog();
            }
            else
            {
                frmACO.Activate();
            }
        }

        /// <summary>
        /// start  a thread to calculate  using ACO Algorithm 
        /// </summary>
        public void BenginACOThread()
        {
            Thread thdACO = new Thread(new ThreadStart(BeginACOCalcate));
            thdACO.IsBackground = true;
            thdACO.Start();
        }

        /// <summary>
        ///  start ACO calculate 
        /// </summary>
        private void BeginACOCalcate()
        {
            //MessageBox.Show("test");

            //CACO cACO = new CACO(CAppParamter.im, CAppParamter.iCityCount, CAppParamter.dAlpha, CAppParamter.dBeta, CAppParamter.dRho,
            //                  CAppParamter.dQ, CAppParamter.iNcMax, xCoordinateArray, yCoordinateArray, CAppParamter.dC, ref  pictureBoxTSP, ref labelCurrentLength, ref bestRoute, this);
            CACO cACO = new CACO(CAppParamter.im, CAppParamter.iCityCount, CAppParamter.dAlpha, CAppParamter.dBeta, CAppParamter.dRho,
                           CAppParamter.dQ, CAppParamter.iNcMax, xCoordinateArray, yCoordinateArray, CAppParamter.dC, ref bestRoute, this);
            cACO.Init();
            cACO.FindOptimalPath();
        }


        /// <summary>
        /// start  a thread  to calculate  using Greedy algorithm 
        /// </summary>
        public void BeginGreedyThread()
        {
            Thread thdGreedy = new Thread(new ThreadStart(StartGreedyCalculate));
            thdGreedy.IsBackground = true;
            thdGreedy.Start();
        }

        /// <summary>
        /// start calculation of Greedy Algorithm 
        /// </summary>
        private void StartGreedyCalculate()
        {
            CGreedyTSP cGreedyTSP = new CGreedyTSP(CAppParamter.iCityCount, xCoordinateArray, yCoordinateArray,
                                       CAppParamter.iStartCityNo, ref bestRoute, this);
            cGreedyTSP.Init();
            cGreedyTSP.FindOptimalPath();
        }

        public delegate void setProgressInvoke(int num);
        /// <summary>
        /// set the Progress of the cut progressBarSolver
        /// </summary>
        /// <param name="num">current vule ofprogressBarSolver </param>
        public void setProgressBar(int num)
        {
            try
            {
                if (this.InvokeRequired)
                {
                    setProgressInvoke _setInvoke = new setProgressInvoke(setProgressBar);
                    this.Invoke(_setInvoke, new object[] { num });
                }
                else
                {
                    progressBarSolver.Value = num;
                }
            }
            catch (System.Exception ex)
            {

            }
        }


        public delegate void setProgressMaxMin(int iMin, int iMax);
        /// <summary>
        /// set the maximum value and minimum value of progressbar 
        /// </summary>
        /// <param name="iMin">minimum value </param>
        /// <param name="iMax">maximum value</param>
        public void setProgressBarMaxMin(int iMin, int iMax)
        {
            if (this.InvokeRequired)
            {
                setProgressMaxMin _setInvoke = new setProgressMaxMin(setProgressBarMaxMin);
                this.Invoke(_setInvoke, new object[] { iMin, iMax });
            }
            else
            {
                progressBarSolver.Minimum = iMin;
                progressBarSolver.Maximum = iMax;
            }
        }

        private void FrmTSPSolver_Load(object sender, EventArgs e)
        {
            progressBarSolver.Minimum = 0;
            labelCurrentLength.Text = "";
            labelIterativeCount.Text = "";
            toolStripStatusLabel1.Text = DateTime.Now.ToString("dddd,dd/M/yyyy");
            timerTimeSecond.Start();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            // this.Close();
            Application.Exit();
            //Application.ExitThread();
        }

        private void FrmTSPSolver_FormClosing(object sender, FormClosingEventArgs e)
        {
            //  this.Dispose();
            if (MessageBox.Show(" Are you sure to exit？", "", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.No) // 2014/5/25-10:28
            {
                e.Cancel = true;
                return;
            }
        }

        private void timerTimeSecond_Tick(object sender, EventArgs e)
        {
            toolStripStatusLabelTimeSecond.Text = DateTime.Now.ToString("HH:mm:ss:ffff");
        }

        private void greedyAlgorithmToolStripMenuItem_Click(object sender, EventArgs e)
        {
            labelCurrentLength.Text = "";
            labelIterativeCount.Text = "";
            labelAlgorithmType.Text = "Greedy Algorithm progress...";
            if (frmGreedy == null || frmGreedy.IsDisposed)
            {
                frmGreedy = new FrmGreedy(this);
                frmGreedy.ShowDialog();
            }
            else
            {
                frmGreedy.Activate();
            }
        }

        private void aboutTheProramToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (frmAbout == null || frmAbout.IsDisposed)
            {
                frmAbout = new FrmAbout();
                frmAbout.ShowDialog();
            }
            else
            {
                frmAbout.Activate();
            }
        }

        private void closeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void pictureBoxTSP_Resize(object sender, EventArgs e)
        {
            //paintPoint();
            //DrawMinRoadLine(bestRoute);
        }

        private void pictureBoxTSP_Click(object sender, EventArgs e)
        {

        }

        private void label1_Click(object sender, EventArgs e)
        {

        }


    }
}
